
/*
   Programme pc_rm5-Horloge: régleur d'horloge à balancier - chronocomparateur
   Auteur : P. Chour - 2023-07-01 - V1.0
   Matériel : carte Arduino Mega, afficheur Touchscreen 3,5 pouces, signal d'entrée, sortie LED optionnelle
   Programmation bourin en dur. Désolé mais pour le coup, j'avais la femme de faire bien et j'étais pressé.
   V1.1 : correction d'un bug dans le calcul de l'AR jour.
   
*/


#include <limits.h>
#include <EEPROM.h>
#include "AffichageTS35.h"

//#define DEBUG_SERIAL

//
// Constantes, define et variables pour la mesure.
//

#define LEDSig 33         // LED signalant un signal : Pin D33 (usage de la LED facultatif).
#define SignalCapteur 21  // Entrée logique signal provenant du capteur : PIN 21.

#define MoyenneGlissante 20  // Pour un tableau destiné au calcul de la moyenne glissante. On peut procéder autrement et plus rapidement (cf. PC-RM4) mais ici, c'est la solution adoptée !

int64_t DemiDureeLongTerme;    // Pour mesure précise des périodes (sur l'ensemble de la mesure).
int32_t NbDemiDureeLongTerme;  // Nombre de mesures DureeLongTerme
int64_t DemiDureeLongTermeAncien;  // Ancienne durée long terme affichée. Sert à éviter de réafficher la valeur si égale à celle affichée.
int64_t AncienAR;

int64_t DureeTicOuTacGlissante[MoyenneGlissante];
int IDebGlissante, NbGlissante;
int64_t DureeTicOuTacGlissanteAncien;

uint64_t SommeEcartTicTac;  // Pour moyenne de la différence de l'écart entre Tic et Tac (sur l'ensemble de la mesure). Toujours positif
uint64_t NbEcartTicTac;     // Compteur pour calcul de la Moyenne EcartTicTac
uint64_t EcartTicTacAncien;

uint64_t TempsTicTacPrec;  // Valeur du timer précédent
uint64_t TempsTicTac;      // Valeur du timer courant

int64_t DureeTicOuTacPrec;  // Valeur précédente de DureeTicOuTac
int64_t DureeTicOuTac;      // Durée d'un Tic ou d'un Tac.

// Période visée pour l'oscillation du pendule
int64_t PeriodeVisee = 1000;     // Valeur de la période visée en ms par défaut
int64_t DemiPeriodeVisee = 500;  // Valeur de la demi période visée en ms

// variables sous IT
volatile bool ITTic = false;  // Vrai si un Tic ou un Tac a été détecté (génération interruption). Doit être remis à faux par l'utilisateur




// Tableau des boutons


TBouton BoutonAffichageMenu[7] = { { LigneBouton, 20, LargeurCol, "Manuel", BLANC, CouleurFond, ROUGE, BLANC },
                                   { LigneBouton, 60 + LargeurCol, LargeurCol, "Mesure", BLANC, CouleurFond, ROUGE, BLANC },
                                   { LigneBouton, 100 + 2 * LargeurCol, LargeurCol, "Pause", BLANC, CouleurFond, ROUGE, BLANC },
                                   { LigneBouton + 40, 20, LargeurCol, "Affect.", BLANC, CouleurFond, ROUGE, BLANC },
                                   { LigneBouton + 40, 60 + LargeurCol, LargeurCol, "Periode",BLANC, CouleurFond, ROUGE, BLANC },
                                   { LigneBouton + 40, 100 + 2 * LargeurCol, LargeurCol, "Reinit", BLANC, CouleurFond, ROUGE, BLANC },
                                   { -1, -1, 0, "", 0, 0, 0, 0 } };

TBouton BoutonAlerte[2] = { { LigneBouton - 40, 60 + LargeurCol, LargeurCol, "Mesure", CouleurFond, CouleurFond, JAUNE, NOIR },
                            { -1, -1, 0, "", 0, 0, 0, 0 } };


const char Depassementms[] = "> 100s\0";
const char Depassementmn[] = "> 24H00\0";

/*
   Initialisation du programme
*/
void setup() {

#ifdef DEBUG_SERIAL
  Serial.begin(9600);
#endif
  AfficheInit();          // Initialisation de l'écran graphique
  AfficheEcranAccueil();  // Affichage de l'écran d'accueil
  delay(1000);
}

//
// La fonction ci-dessous mesure le temps en prenant en compte un éventuel bouclage de la fonction millis ou micro
// On lui donne la valeur de départ de la valeur à mesurer et la valeur courante de millis ou micro. Elle rend la durée écoulée en millisecondes ou microseconde.
// DureeVrai = courante-dernière
//
unsigned long DureeVrai(unsigned long derniere, unsigned long courante) {
  if (courante < derniere) {  // on est passé par zéro
    return (0xffffffffu - derniere) + courante;
  } else {
    return courante - derniere;
  }
}

//
// Cette fonction transforme un uint en Ascii et ajoute le texte des unités (ms, s...). Le résultat est stocké dans Buf[i]...Buf[i+nb car-1] avec un 0 à la fin.
// i en sortie pointe sur le 0. On peut donc enchainer l'appel à cette fonction. Dans le cas présent, permet de construire quelque chose du style 12mn3s12ms.
//

void ConvIntCharEtConcatene(uint64_t V1, int *i, const char Unite[], char Buf[]) {
  char BufTemp[MaxCarBuf];
  int L, j;
  L = 0;
  do {
    BufTemp[L] = (V1 - (V1 / 10) * 10) + 48;
    V1 = V1 / 10;
    L = L + 1;
  } while (V1 > 0);
  while (L > 0) {
    Buf[*i] = BufTemp[L - 1];
    L = L - 1;
    *i = *i + 1;
  }
  j = 0;
  while (Unite[j] != '\0') {
    Buf[*i] = Unite[j];
    *i = *i + 1;
    j = j + 1;
  }
  Buf[*i] = '\0';
}

//
// Formatte une valeur temporelle pour affichage. Résultat dans un tableau de char terminé par null.
// V1 = valeur en microsecondes (précision de l'Arduino)
// Signe = vrai : le signe ets affiché. Sinon, on met un espace.
// MS = vrai si résultat en milliseconde seulement. MS = faux si résultat en minutes, secondes et millisecondes
// Rend un message en clair si débordement (valeur en ms > 100000 ou valeur en mn > 1440 (24Heures))
// Pas de vérification de débordement !!! L'appelant doit faire attention que buf ait la bonne taille.
//
void FormatTemps(int64_t V1, bool Signe, char buf[], bool MS) {
  uint64_t mn, ss, ms;
  int L, i;
  char BufTemp[MaxCarBuf];
  char CarSigne = ' ';
  int iPF, k, j;
  if (L > MaxCarBuf) { L = MaxCarBuf; }  // Une précaution !
  // traitement du signe avec initialisation de CarSigne et V1 = |V1|
  if (Signe) {
    if (V1 > 0) {
      CarSigne = '+';
    } else {
      CarSigne = '-';
    }
  }
  if (V1 < 0) { V1 = -V1; }
  V1 = V1 / 1000;  // On ramène tout aux millisecondes. Suffisant pour une horloge
  if (MS) {        // si résultat en ms, alors
    mn = 0;
    ss = 0;
    ms = V1;
    if (ms > 100000) {
      memcpy(buf, Depassementms, sizeof(Depassementms));
      return;
    }
    i = 1;
    ConvIntCharEtConcatene(ms, &i, "ms\0", buf);
  } else {  // si résultat en unité adaptées
    mn = V1 / 60000;
    V1 = (V1 - (mn * 60000));
    ss = V1 / 1000;
    ms = V1 - (ss * 1000);
    if (mn > 1440) {
      memcpy(buf, Depassementmn, sizeof(Depassementmn));
      return;
    }
    i = 1;
    if (mn > 0) {
      ConvIntCharEtConcatene(mn, &i, "mn\0", buf);
      ConvIntCharEtConcatene(ss, &i, "s\0", buf);
    } else {
      if (ss > 0) {
        ConvIntCharEtConcatene(ss, &i, "s\0", buf);
      }
    }
    ConvIntCharEtConcatene(ms, &i, "ms\0", buf);
  }
  buf[0] = CarSigne;
}

//--------------------------------------------------------------------------------------
// Affichage des mesures
// AAfficher : 0 = Différence TicTac ou amplitude, 1=Différence période ou T3-T1, 2= moyenne période, 3=avance retard par jour, 4 = Période visée
// Force = true : affichage même si pas de différence avec ancien affichage. False : n'affiche que s'il y a des différences
// Tout = true => Force et on affiche tout 
//--------------------------------------------------------------------------------------
void AffMesure(uint8_t AAfficher, bool Force, bool Tout) {
  char buf[MaxCarBuf];
  if (Tout) {
    Force = true;
    AAfficher = 0;
  }
  // Affichage période. On n'affiche que si l'on a au moins deux mesures
  switch (AAfficher) {

    // Moyenne de la période sur toute la durée de la mesure
    case 0:
      {
        int64_t VTemp = 0;
        if (NbDemiDureeLongTerme > 0) {
          VTemp = (DemiDureeLongTerme / NbDemiDureeLongTerme) ;
          if ((VTemp != DemiDureeLongTermeAncien) || Force) {  // on vérifie si ça vaut le coup d'afficher
            FormatTemps(VTemp * 2, false, buf, true);
            AfficheBloc(0, CouleurFond, VERT, buf);
            DemiDureeLongTermeAncien = VTemp;
          }
        } else {
          AfficheBloc(0, CouleurFond, VERT, "-  ");
        }
      }
      if (!Tout) break;

    // Moyenne de la mesure glissante
    case 1:
      {
        if (NbGlissante > 0) {
          int64_t VTemp = 0;
          for (int i = 0; i < NbGlissante; i++) { VTemp = VTemp + DureeTicOuTacGlissante[i]; }
          VTemp = VTemp / NbGlissante;
          if ((VTemp != DureeTicOuTacGlissanteAncien) || Force) {  // on vérifie si ça vaut le coup d'afficher
            FormatTemps(VTemp * 2, false, buf, true);
            AfficheBloc(1, CouleurFond, VERT, buf);
            DureeTicOuTacGlissanteAncien = VTemp;
          }
        } else {
          AfficheBloc(1, CouleurFond, VERT, "-  ");
        }
      }
      if (!Tout) break;

    // Moyenne des différences entre Tic et Tac ou affichage Amplitude selon type d'affichage
    case 2:
      {
        uint64_t VTemp = 0;
        if (NbEcartTicTac > 0) { 
          VTemp = (SommeEcartTicTac / NbEcartTicTac); 
          if ((VTemp != EcartTicTacAncien) || Force) {  // on vérifie si ça vaut le coup d'afficher
            EcartTicTacAncien = VTemp;
            FormatTemps(VTemp, false, buf, true);  // toujours positif
            AfficheBloc(2, CouleurFond, VERT, buf);
          }
        } else {
          AfficheBloc(2, CouleurFond, VERT, "-  ");
        }          
      }
      if (!Tout) break;

    //Avance retard par jours
    case 3:
      {
        int64_t VTemp = 0;
        if (NbDemiDureeLongTerme > 0) { 
          VTemp = DemiPeriodeVisee - (DemiDureeLongTerme / NbDemiDureeLongTerme) ; 
          if ((VTemp != AncienAR) || Force) {
            AncienAR = VTemp;
            FormatTemps(2 * VTemp * 86400000000 / PeriodeVisee, true, buf, false); // 86 400 000 000 Nb de microsecondes dans une journée
            AfficheBloc(3, CouleurFond, VERT, buf);
          }
        } else {
          AfficheBloc(3, CouleurFond, VERT, "-  ");
        }
      }
      if (!Tout) break;

    //Période Visée
    case 4:
      {
        FormatTemps(PeriodeVisee, false, buf, true);
        AfficheBloc(4, CouleurFond, JAUNE, buf);
      }
      break;      
    default:;
  }
}


/*
 * Changement état Pin Signal sur IT. 
 * fonction déclenchée par IT sur changement d'état de SignalCapteur
 */
void SignalChange() {
  unsigned long Temp;
  Temp = micros();                                   // On sauve le temps courant en microsecondes
  digitalWrite(LEDSig, digitalRead(SignalCapteur));  // On allume ou on éteint la LED selon la valeur du signal
  if (!ITTic) {                                      // Si IT en cours de traitement, on l'ignore
    if (digitalRead(SignalCapteur)) {                // Si la valeur est haute, on suppose qu'il y a eu un front montant. On traite la suite
      TempsTicTac = Temp;                            // Le temps en entrée est stockée dans TempsTicTac
      ITTic = true;                                  //On positionne IITic pour signaler qu'il y a quelque chose à traiter
    }
  }
}

/*
 * Réinitialisation des valeurs de mesure et de l'affichage.
 */

void Reinit() {
  char BufTemp[MaxCarBuf];
  DureeTicOuTacPrec = 0;
  EcartTicTacAncien = 0;
  SommeEcartTicTac = 0;
  NbEcartTicTac = 0;
  DemiDureeLongTerme = 0;
  NbDemiDureeLongTerme = 0;
  AncienAR = 0;
  IDebGlissante = 0;
  NbGlissante = 0;
  DureeTicOuTacGlissanteAncien = 0;
  AffMesure(0, true, true);
}

/*
 * APPLICATION
 * Pour une fois, je n'utilise pas d'automates d'états finis !
 */

void loop() {
  int i;
  byte AAfficher = 0;      // pour affichage des mesures. Pour des raisons de temps d'affichage, on n'affiche qu'une mesure à la fois
  unsigned long AffTimer;  // On ne fait un affichage des mesures que toutes les 1000ms (Arduino, c'est lent !)
  unsigned long V1;

  bool MesureAlerte = false;       // Pour signaler qu'il y a un problème lors de la mesure
  bool AttentePremierTic = false;  // Attente d'un premier TIC
  bool EtatMesure = false;         // on mesure les écarts et la période
  bool EtatPeriode = false;

  char BufTemp[MaxCarBuf];
  int BoutonPresse = -1;  // Vaut la valeur du bouton pressé, sinon, -1
  int IndexMenu;          // Valeur courante de l'index du menu
  const unsigned int AdIndexPeriode = 0;  // Adresse de la sauvegarde de la période visée en EEPROM

  EEPROM.get(AdIndexPeriode, PeriodeVisee);                    // Récupération des valeurs par défaut de IndexBattements, IndexOccultation et supr
  if (PeriodeVisee >= 3600000000) { PeriodeVisee = 1000000; }  // Période visée > 1 heure, on force à 1 seconde
  DemiPeriodeVisee = PeriodeVisee / 2;

  pinMode(LEDSig, OUTPUT);     // LED Signalant un battement
  digitalWrite(LEDSig, HIGH);  // On allume la LED
  delay(500);                  // Pendant 500ms
  digitalWrite(LEDSig, LOW);   // Puis on l'éteint
  pinMode(SignalCapteur, INPUT);
  attachInterrupt(digitalPinToInterrupt(SignalCapteur), SignalChange, CHANGE);  // Activation de l'IT sur changement d'état du signal provenant du capteur

  AffTimer = millis();  // Sert pour déclencher les affichages des mesures

  IndexMenu = 2;                                           // le bouton pause a pour index 2
  AfficheEcranMenu(BoutonAffichageMenu);                   // on affiche l'écran principal
  AfficheUnBouton(&BoutonAffichageMenu[IndexMenu], true);  // Et on indique qu'on est en pause
  Reinit();

  while (1) {
    if (IsPressed()) {                                                // A-t-on appuyé sur l'écran ?
      BoutonPresse = NumBouton(BoutonAffichageMenu);                  // on récupère la valeur du bouton
      if (BoutonPresse != -1) {                                       // On teste la validité du  bouton. Si Ok, on continue
        if (BoutonPresse != IndexMenu) {                              // Si le bouton pressé est différent du dernier bouton pressé
          AfficheUnBouton(&BoutonAffichageMenu[IndexMenu], false);    // Alors, on désactive le dernier bouton pressé
          ClignoteBouton(&BoutonAffichageMenu[BoutonPresse]);         // On fait clignoter le nouveau bouton
          AfficheUnBouton(&BoutonAffichageMenu[BoutonPresse], true);  // Et on l'active
        }
        switch (BoutonPresse) {  // Selon le bouton pressé
          case 0:                // Saisie manuelle de la période visée.
            V1 = PeriodeVisee/1000;
            if (SaisieClavier(&V1) && (V1 != 0)) {
              PeriodeVisee = V1*1000;
              EEPROM.put(AdIndexPeriode, PeriodeVisee);  
              DemiPeriodeVisee = PeriodeVisee/2;
              DemiDureeLongTerme = 0; // Les écarts longs ne sont plus valides. On les réinitialise
              NbDemiDureeLongTerme = 0;
              DemiDureeLongTermeAncien = 0;
            }
            AfficheEcranMenu(BoutonAffichageMenu);  // Réaffichage de l'écran de mesure
            AfficheUnBouton(&BoutonAffichageMenu[BoutonPresse], true);
            AffMesure(0,true,true);
            break;
          case 4:                      // Période.
          case 1:                      // Mesure.
            AttentePremierTic = true;  // On attend le premier Tic
            ITTic = false;             // Et par précaution, on oublie l'éventuelle IT en cours non traitée
            break;
          case 2:  // Passage en pause
            break;
          case 3:  // Affectation Période moyenne à période visée. Il est prudent après coup de faire un réinit !
            if (NbDemiDureeLongTerme > 0) { 
              V1 = (DemiDureeLongTerme / NbDemiDureeLongTerme); 
              if (V1 > 0) {
                PeriodeVisee = V1*2;
                DemiPeriodeVisee = V1;
                EEPROM.put(AdIndexPeriode, PeriodeVisee);  
                DemiDureeLongTerme = 0;
                NbDemiDureeLongTerme = 0;
                DemiDureeLongTermeAncien = 0;
                AffMesure(0,true,true);
              }
            }
            break;
          case 5:  // Réinit
            Reinit();
            break;
        }
        IndexMenu = BoutonPresse;
        EtatPeriode = (IndexMenu == 4);
        EtatMesure = (IndexMenu == 1);
        BoutonPresse = -1;
      }
    }

    // Attente d'un premier Tic
    if (AttentePremierTic) {
      if (ITTic) {                      // on l'a eu
        TempsTicTacPrec = TempsTicTac;  // On conserve l
        AttentePremierTic = false;
        EtatPeriode = (IndexMenu == 4);
        EtatMesure = (IndexMenu == 1);
        ITTic = false;
      }
    }

    // Dans cet état, on mesure la durée d'une période et les écarts entre deux demis période. Lors de la mesure de la seule période, on ne prend pas en compte la valeur de la période visée. Dans l'autre cas (mesure de l'A/R)
    // on considère que si la demi période mesurée est > la période visée, alors il y a un problème et on ne prend pas en compte la valeur. Par contre, on affiche une erreur
    if (EtatPeriode || EtatMesure) {
      if (ITTic) {  // Si on arrive dans cet état, c'est qu'on a au moins un TicTacPrec. On détermine la valeur du TicTacCourant
        DureeTicOuTac = DureeVrai(TempsTicTacPrec, TempsTicTac);
        TempsTicTacPrec = TempsTicTac;
        if ((DureeTicOuTac <= PeriodeVisee) || (EtatPeriode)) {
          if (MesureAlerte) {
            AfficheUnBouton(&BoutonAlerte[0], false);
            MesureAlerte = false;
          }
          if (DureeTicOuTacPrec != 0) {

            // Calcul de la moyenne de la différence entre un Tic et un Tac. La valeur est toujours positive. Comme elle s'accumule, on prend un 64 bits.
            if (DureeTicOuTacPrec > DureeTicOuTac) {
              SommeEcartTicTac = SommeEcartTicTac + DureeTicOuTacPrec - DureeTicOuTac;
            } else {
              SommeEcartTicTac = SommeEcartTicTac + DureeTicOuTac - DureeTicOuTacPrec;
            }
            NbEcartTicTac = NbEcartTicTac + 1;
          }

          // pour le calcul de la moyenne de la demi-période sur le long terme
          DemiDureeLongTerme = DemiDureeLongTerme + DureeTicOuTac;
          NbDemiDureeLongTerme = NbDemiDureeLongTerme + 1;

          // pour le calcul de la moyenne de la demi-période sur une période glissante
          DureeTicOuTacGlissante[IDebGlissante] = DureeTicOuTac;
          IDebGlissante = IDebGlissante + 1;
          if (IDebGlissante >= MoyenneGlissante) { IDebGlissante = 0; }
          if (NbGlissante < MoyenneGlissante) { NbGlissante = NbGlissante + 1; }

          DureeTicOuTacPrec = DureeTicOuTac;

          // Affichages toutes les secondes. On n'affiche pas l'avance-retard et on affiche qu'une valeur à la fois car ça prend du temps ! La valeur à afficher change à chaque appel.
          if (DureeVrai(AffTimer, millis()) > 1000) {
            AffTimer = millis();
            AffMesure(AAfficher, false, false);
            if (EtatPeriode) {
              AAfficher = (AAfficher + 1) % 3;
            } else {
              AAfficher = (AAfficher + 1) % 4;
            }
          }
        } else {
          if (!MesureAlerte) {
            AfficheUnBouton(&BoutonAlerte[0], true);
            MesureAlerte = true;
          }
        }
      }
      ITTic = false;
    }
  }
}
